use crate::{
    instructions,
    stack::{Instruction, Stack, StackVec},
};

fn find_labels(stack: &mut Stack) {
    while stack.program_counter < (stack.program.len() as u16) {
        let label = &stack.program[stack.program_counter as usize];
        stack.program_counter += 1;

        if !label.name.starts_with('@') {
            continue;
        }

        if stack.labels[label.data as usize].is_some() {
            eprintln!("Label {} already defined", label.data);
            std::process::exit(2);
        }

        if label.data == 0 {
            stack.labels[0] = Some(stack.program_counter as i16);
            continue;
        }

        stack.labels[label.data as usize] = Some(stack.program_counter as i16);
    }

    stack.program_counter = stack.labels[0].unwrap_or(0) as u16;
}

fn find_prefix_operators(instruction: &mut Instruction, memory: &mut StackVec) {
    match instruction.name.chars().nth(0) {
        Some('|') => {
            instruction.name = instruction.name.chars().skip(1).collect();
            instruction.data = memory.pop();
        }
        Some('^') => {
            instruction.name = instruction.name.chars().skip(1).collect();

            if instruction.name.chars().nth(0) != Some('#') {
                println!("{0} {1}", instruction.name, instruction.data.to_string());
            } else if instruction.arg.is_empty() {
                println!("{}", instruction.name);
            } else {
                println!("{0} {1}", instruction.name, instruction.arg);
            }
        }
        _ => {}
    }
}

pub fn execute(stack: &mut Stack, mut origin_stack: Option<&mut Stack>, mod_name: String) {
    find_labels(stack);

    while let Some(mut instruction) = stack.program.get_mut(stack.program_counter as usize) {
        stack.program_counter += 1;

        find_prefix_operators(&mut instruction, &mut stack.memory);

        if instruction.name.starts_with('@') {
            continue;
        }

        match instruction.name.as_str() {
            // Stack operations
            "push" => instructions::stack_manage::push(&mut stack.memory, instruction.data),
            "pop" => instructions::stack_manage::pop(&mut stack.memory),
            "dup" => instructions::stack_manage::dup(&mut stack.memory),
            "swap" => instructions::stack_manage::swap(&mut stack.memory),
            "pick" => instructions::stack_manage::pick(&mut stack.memory, instruction.data),

            // Math operations
            "add" => instructions::math::add(&mut stack.memory),
            "sub" => instructions::math::sub(&mut stack.memory),
            "mul" => instructions::math::mul(&mut stack.memory),
            "div" => instructions::math::div(&mut stack.memory),
            "mod" => instructions::math::modulo(&mut stack.memory),

            // Logic
            "and" => instructions::logic::and(&mut stack.memory),
            "or" => instructions::logic::or(&mut stack.memory),
            "xor" => instructions::logic::xor(&mut stack.memory),
            "nand" => instructions::logic::nand(&mut stack.memory),
            "not" => instructions::logic::not(&mut stack.memory),

            // Bitwise
            "band" => instructions::bitwise::band(&mut stack.memory),
            "bor" => instructions::bitwise::bor(&mut stack.memory),
            "bxor" => instructions::bitwise::bxor(&mut stack.memory),
            "bnand" => instructions::bitwise::bnand(&mut stack.memory),
            "bnot" => instructions::logic::not(&mut stack.memory),

            // Flow-control
            "jmp" => instructions::flow_control::jmp(
                &mut stack.labels,
                &mut stack.program_counter,
                instruction.data,
            ),
            "rjmp" => instructions::specific::rjmp(
                &mut stack.memory,
                &mut stack.labels,
                &mut stack.program_counter,
            ),
            "jnz" => instructions::flow_control::jnz(
                &mut stack.memory,
                &mut stack.labels,
                &mut stack.program_counter,
                instruction.data,
            ),
            "jz" => instructions::flow_control::jz(
                &mut stack.memory,
                &mut stack.labels,
                &mut stack.program_counter,
                instruction.data,
            ),
            "kjnz" => instructions::flow_control::kjnz(
                &mut stack.memory,
                &mut stack.labels,
                &mut stack.program_counter,
                instruction.data,
            ),
            "kjz" => instructions::flow_control::kjz(
                &mut stack.memory,
                &mut stack.labels,
                &mut stack.program_counter,
                instruction.data,
            ),

            // Input/Output
            "putc" => instructions::io::putc(&mut stack.memory),
            "getc" => instructions::io::getc(&mut stack.memory),
            "scan" => instructions::io::scan(&mut stack.memory),
            "meow" => instructions::io::meow(&mut stack.memory),
            "#print" => instructions::io::print(instruction.arg.clone()),
            "$print" => instructions::io::print(instruction.arg.clone()),
            "#println" => instructions::io::println(instruction.arg.clone()),
            "$println" => instructions::io::println(instruction.arg.clone()),

            // Modules
            "args" => {
                instructions::modules::args(&mut stack.memory, &mut origin_stack, instruction.data)
            }
            "#exec" => {
                let arg = instruction.arg.clone();
                instructions::modules::exec(stack, arg, mod_name.clone())
            }

            // Self-modifying
            "#insert" => {
                let arg = instruction.arg.clone();
                let data = instruction.data.clone();
                instructions::self_modify::insert(&mut stack.program, arg, data)
            }

            // Special
            "dump" => instructions::special::dump(&mut stack.memory),
            "size" => instructions::special::size(&mut stack.memory),
            "maxsize" => instructions::special::maxsize(&mut stack.memory),
            "#expr" => instructions::special::expr(&mut stack.memory, instruction.arg.clone()),
            "nop" => continue,
            "quit" => break,
            "exit" => instructions::special::exit(instruction.data),

            // Platform-specific
            #[cfg(target_family = "unix")]
            "_unix_random" => instructions::platform_specific::_unix_random(&mut stack.memory),

            #[cfg(not(target_family = "unix"))]
            "_unix_random" => instructions::specific::random(&mut stack.memory),

            // Labast-specific
            "random" => instructions::specific::random(&mut stack.memory),

            _ => {
                eprintln!("Unknown instruction: {}", instruction.name);
                std::process::exit(2);
            }
        }
    }
}
